@@ -335,12 +335,13 @@ def __get_stations(ids=[], progress=True):
     # dictionary to return
     stations = {}
 
-    # fill dictionary with ICOS station id, latitude, longitude and altitude
-	 # implement progress True/False
-    for ist in tqdm(sorted(allStations), disable=progress):
-        if not ist in df['STILT id'].values:
-            continue
+    # load precomputed geoinfo
+    geo = geoinfo.get()
 
+    # fill dictionary with ICOS station id, latitude, longitude and altitude
+    # implement progress True/False
+    for ist in tqdm(sorted(allStations), disable=progress):
+        
         stations[ist] = {}
         # get filename of link (original stiltweb directory structure) and
         # extract location information
@@ -362,8 +363,16 @@ def __get_stations(ids=[], progress=True):
         stations[ist]['locIdent']=os.path.split(loc_ident)[-1]
 
         # set the name and id
-        stations[ist]['id'] = df.loc[df['STILT id'] == ist]['STILT id'].item()
-        stationName = str(df.loc[df['STILT id'] == ist]['STILT name'].item())
+        stations[ist]['id'] = ist
+        
+        # check if station id is in the manual curated list at
+        # CPC.STILTINFO
+        id = [a.lower() for a in df['STILT id'].values]
+        if ist.lower() in id:
+            stationName = str(df.loc[df['STILT id'] == ist]['STILT name'].item())
+        else:
+            stationName = ''
+        
         stations[ist]['name'] = __stationName(stations[ist]['id'], stationName, stations[ist]['alt'])
 
         # set a flag if it is an ICOS station
@@ -384,15 +393,18 @@ def __get_stations(ids=[], progress=True):
             months = sorted([m for m in months if not sub.lower() in m.lower()])
             stations[ist][yy]['months'] = months
             stations[ist][yy]['nmonths'] = len(stations[ist][yy]['months'])
-
-    # merge geoinfo
-    geo = geoinfo.get()
-
-    for k in stations:
-        if k in geo.keys():
-            stations[k]['geoinfo'] = geo[k]['geoinfo']
+            
+        # add geoinfo
+        if ist in geo.keys():
+            #get the precomputed country info
+            stations[ist]['geoinfo'] = geo[ist]['geoinfo']
+        elif stations[ist]['icos']:
+            # get country from ICOS coordinates
+            stations[ist]['geoinfo'] = __country([stations[ist]['icos']['lat'],stations[ist]['icos']['lon']])
         else:
-            stations[k]['geoinfo'] = __country([stations[k]['lat'],stations[k]['lon']])
+            # get country from STILT coordiantes
+            stations[ist]['geoinfo'] = __country([stations[ist]['lat'],stations[ist]['lon']])
+            
 
     return stations
 
@@ -404,7 +416,7 @@ def __country(latlon):
 
 
 def __stationName(idx, name, alt):
-    if name=='nan':
+    if name=='nan' or not name:
         name = idx
     if not (name[-1]=='m' and name[-2].isdigit()):
         name = name + ' ' + str(alt) + 'm'
@@ -539,7 +551,6 @@ def find(**kwargs):
         del kwargs['edate']
 
 
-
     # valid key words. Make sure all are lower capital and that
     # the function has been defined above and that outfmt is the very
     # last call:
